import pytest
import numpy as np
import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import modelautopsy

def test_empty_tensor():
    with pytest.raises(RuntimeError):
        modelautopsy.analyze(np.array([], dtype=np.float32))

def test_healthy_tensor():
    data = np.random.randn(1000).astype(np.float32)
    assert not modelautopsy.has_failure(data)
    
    report = modelautopsy.analyze(data)
    assert report["nan_count"] == 0
    assert report["inf_count"] == 0
    assert report["valid_count"] == 1000

def test_nan_detection():
    data = np.array([1.0, 2.0, np.nan, 4.0], dtype=np.float32)
    assert modelautopsy.has_failure(data) == True
    
    report = modelautopsy.analyze(data)
    assert report["nan_count"] == 1
    assert report["valid_count"] == 3

def test_inf_detection():
    data = np.array([1.0, np.inf, -np.inf], dtype=np.float32)
    assert modelautopsy.has_failure(data) == True
    
    report = modelautopsy.analyze(data)
    assert report["inf_count"] == 2
    # Mean should ignore inf
    assert report["mean"] == 1.0 

def test_mixed_failures():
    # Large dataset test
    size = 1000000
    data = np.random.randn(size).astype(np.float32)
    data[0] = np.nan
    data[500] = np.inf
    data[-1] = np.nan
    
    assert modelautopsy.has_failure(data) == True
    report = modelautopsy.analyze(data)
    
    assert report["nan_count"] == 2
    assert report["inf_count"] == 1
    assert report["valid_count"] == size - 3

def test_wrong_dimensions():
    # 2D tensor should fail or require flattening
    data = np.random.randn(10, 10).astype(np.float32)
    with pytest.raises(RuntimeError):
        modelautopsy.analyze(data)

if __name__ == "__main__":
    pytest.main([__file__, "-v"])


import pytest
import numpy as np
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import modelautopsy
from debugger import watch, _prepare_tensor

class TestCoreEngine:
    def test_empty_tensor(self):
        """Should handle size 0 without crash."""
        data = np.array([], dtype=np.float32)
        report = modelautopsy.analyze(data)
        assert report["nan_count"] == 0
        assert report["mean"] == 0.0

    def test_all_nans(self):
        """Pure NaN stress test."""
        data = np.full(100, np.nan, dtype=np.float32)
        report = modelautopsy.analyze(data)
        assert report["nan_count"] == 100
        assert report["valid_count"] == 0
        assert report["mean"] == 0.0

    def test_float64_conversion(self):
        """Wrapper should handle float64 inputs."""
        # Note: This tests the python wrapper _prepare_tensor
        data = np.random.randn(100).astype(np.float64) # Wrong type
        prepared = _prepare_tensor(data)
        assert prepared is not None
        assert prepared.dtype == np.float32
        assert prepared.size == 100

    def test_list_input(self):
        """Wrapper should handle Python lists."""
        data = [1.0, 2.0, np.nan]
        prepared = _prepare_tensor(data)
        assert prepared is not None
        assert prepared.dtype == np.float32
        assert np.isnan(prepared[2])

    def test_mixed_types(self):
        """Test handling of integers and floats."""
        data = [1, 2, 3.5]
        prepared = _prepare_tensor(data)
        assert prepared is not None
        assert prepared.dtype == np.float32

class TestDecorator:
    def test_decorator_no_failure(self):
        @watch(drop_into_debugger=False, verbose=False)
        def clean_func(x):
            return x * 2
        
        result = clean_func(np.array([1.0, 2.0]))
        assert result is not None

    def test_decorator_detection(self):
        @watch(drop_into_debugger=False, verbose=False)
        def corrupt_func(x):
            x[0] = np.nan
            return x

        # Mock print to capture output logic if needed
        result = corrupt_func(np.array([1.0, 2.0]))
        assert result is not None # Function runs
        # Ideally we'd capture logs, but for unit test, just ensure it doesn't crash

    def test_decorator_non_tensor_args(self):
        @watch(drop_into_debugger=False, verbose=False)
        def mixed_func(x, y):
            return x + y
        
        # Should not crash on non-array args
        res = mixed_func(5, 10)
        assert res == 15

if __name__ == "__main__":
    pytest.main([__file__, "-v", "--tb=short"])


"""
ModelAutopsy: Enterprise ML Debugging Library.
"""
from .debugger import watch
from .debugger import analyze as _internal_analyze

# Engine Detection
ENGINE_STATUS = "Unknown"

try:
    # Try to import Rust module
    import modelautopsy_rust
    ENGINE_STATUS = "Rust Safety Core"
except ImportError:
    # Try C++ Module
    try:
        import _core_cpp
        ENGINE_STATUS = "C++ Speed Core"
    except ImportError:
        ENGINE_STATUS = "Pure Python (Slow)"

# Expose Analyze
def analyze(tensor):
    """Public API. Analyzes tensor using best available engine."""
    return _internal_analyze(tensor)

__all__ = ["watch", "analyze", "ENGINE_STATUS"]



"""
IDE Debugger.
"""
import functools
import inspect
import numpy as np
from rich.console import Console
from rich.table import Table

console = Console()

# 1. Load Engines
try:
    import modelautopsy_rust
    RUST_AVAILABLE = True
except ImportError:
    RUST_AVAILABLE = False

# 2. Load C++ Engine
try:
    import _core_cpp
    CPP_AVAILABLE = True
except ImportError:
    CPP_AVAILABLE = False

def _prepare_tensor(tensor):
    if not isinstance(tensor, np.ndarray):
        try:
            tensor = np.array(tensor)
        except:
            return None
    if tensor.dtype != np.float32:
        tensor = tensor.astype(np.float32)
    if tensor.ndim != 1:
        tensor = tensor.flatten()
    return tensor

def _convert_rust_to_dict(rust_obj):
    return {
        "nan_count": rust_obj.nan_count,
        "inf_count": rust_obj.inf_count,
        "valid_count": rust_obj.valid_count,
        "mean": rust_obj.mean,
        "variance": rust_obj.variance,
        "l2_norm": rust_obj.l2_norm,
        "min_val": rust_obj.min_val,
        "max_val": rust_obj.max_val,
    }

def analyze(tensor):
    """Performs C++ scan safely."""
    tensor = _prepare_tensor(tensor)
    if tensor is None: return None

    if RUST_AVAILABLE:
        rust_obj = modelautopsy_rust.rust_analyze(tensor)
        return _convert_rust_to_dict(rust_obj)

    if CPP_AVAILABLE:
        return _core_cpp.analyze(tensor)

    console.print("[red]Warning: Using Pure Python (Slow)[/red]")
    return {
        "nan_count": np.isnan(tensor).sum(),
        "inf_count": np.isinf(tensor).sum(),
        "valid_count": tensor.size - np.isnan(tensor).sum() - np.isinf(tensor).sum(),
        "mean": float(np.mean(tensor)) if tensor.size > 0 else 0,
        "variance": float(np.var(tensor)) if tensor.size > 0 else 0,
        "l2_norm": float(np.linalg.norm(tensor)),
        "min_val": float(np.min(tensor)) if tensor.size > 0 else 0,
        "max_val": float(np.max(tensor)) if tensor.size > 0 else 0,
    }

def _log_error(func, source, report):
    console.print(f"\n[bold red]CRITICAL FAILURE[/bold red]")
    t = Table(show_header=False, header_style="bold magenta")
    t.add_row("[dim]Function:[/dim]", f"[bold]{func}[/bold]")
    t.add_row("[dim]Source:[/dim]", f"[bold]{source}[/bold]")
    t.add_row("[dim]NaN Count:[/dim]", f"[red]{report['nan_count']}[/red]")
    t.add_row("[dim]Inf Count:[/dim]", f"[red]{report['inf_count']}[/red]")
    t.add_row("[dim]L2 Norm:[/dim]", f"{report['l2_norm']:.4e}")
    t.add_row("[dim]Variance:[/dim]", f"{report['variance']:.4e}")
    console.print(t)

def _log_warning(func, source, report):
    console.print(f"\n[bold yellow]TRAP DETECTED[/bold yellow]")
    t = Table(show_header=False)
    t.add_row("[dim]Function:[/dim]", f"[bold]{func}[/bold]")
    t.add_row("[dim]Source:[/dim]", f"[bold]{source}[/bold]")
    t.add_row("[dim]NaN Count:[/dim]", f"[yellow]{report['nan_count']}[/yellow]")
    t.add_row("[dim]Inf Count:[/dim]", f"[yellow]{report['inf_count']}[/yellow]")
    t.add_row("[dim]Action:[/dim]", f"[green]Continued[/green]")
    console.print(t)

def watch(drop_into_debugger=False, verbose=True, inspect_args=True, inspect_return=True):
    """
    Enhanced Decorator.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if verbose:
                console.print(f"[dim]Execution started: {func.__name__}[/dim]")

            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            
            failure_detected = False
            
            # Scan Inputs
            if inspect_args:
                for name, val in bound_args.arguments.items():
                    rep = analyze(val)
                    if rep and (rep['nan_count'] > 0 or rep['inf_count'] > 0):
                        _log_error(func.__name__, f"Input '{name}'", rep)
                        failure_detected = True

            # Execute
            result = func(*args, **kwargs)

            # Scan Output
            if inspect_return:
                rep = analyze(result)
                if rep and (rep['nan_count'] > 0 or rep['inf_count'] > 0):
                    _log_error(func.__name__, "Return Value", rep)
                    failure_detected = True
            
            if verbose and not failure_detected:
                console.print(f"[green]SUCCESS[/green] [dim]{func.__name__} finished clean.[/dim]")

            if failure_detected and drop_into_debugger:
                console.print("[bold red]Halting execution...[/bold red]")
                import pdb; pdb.set_trace()

            return result
        return wrapper
    return decorator



from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup, find_packages
import sys

if sys.platform == "win32":
    copt, lopt = ['/O2', '/std:c++17', '/openmp'], []
else:
    copt, lopt = ['-O3', '-march=native', '-std=c++17', '-fopenmp'], ['-fopenmp']

ext_modules = [
    Pybind11Extension(
        "_core_cpp",
        sources=["src/module.cpp", "src/stats.cpp"],
        include_dirs=["include"],
        extra_compile_args=copt,
        extra_link_args=lopt,
    ),
]

setup(
    name="modelautopsy",
    version="1.0.5", # <--- INCREMENT VERSION TO BREAK LOCK
    author="Yuvraj Jha",
    description="ModelAutopsy: High-Performance ML Failure Detection",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/YuvrajJha13/MLGuardian",
    packages=find_packages(exclude=("tests", "examples")),
    ext_modules=ext_modules,
    install_requires=["numpy", "rich"],
    setup_requires=["pybind11"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.8",
    zip_safe=False,
)
